@@ -53,22 +53,36 @@ module String = struct
   module Map = Map.Make (String)
 
   let is_suffix t ~suffix = Filename.check_suffix t suffix
+
+  let is_prefix t ~prefix =
+    let len_s = length t
+    and len_pre = length prefix in
+    let rec aux i =
+      if i = len_pre
+      then true
+      else if unsafe_get t i <> unsafe_get prefix i
+      then false
+      else aux (i + 1)
+    in
+    len_s >= len_pre && aux 0
+  ;;
 end
 
 module List = struct
   include List
 
   let partition_map_skip t ~f =
-    let rec loop l r = function
-      | [] -> l, r
+    let rec loop l m r = function
+      | [] -> l, m, r
       | x :: xs ->
         (match f x with
-         | `Skip -> loop l r xs
-         | `Left x -> loop (x :: l) r xs
-         | `Right x -> loop l (x :: r) xs)
+         | `Skip -> loop l m r xs
+         | `Left x -> loop (x :: l) m r xs
+         | `Middle x -> loop l (x :: m) r xs
+         | `Right x -> loop l m (x :: r) xs)
     in
-    let l, r = loop [] [] t in
-    rev l, rev r
+    let l, m, r = loop [] [] [] t in
+    rev l, rev m, rev r
   ;;
 
   let rec filter_map l ~f =
@@ -114,6 +128,7 @@ module Status_line = struct
       displayed := new_displayed)
   ;;
 
+  let clear () = Printf.printf "\r*s\r%!"
   let () = at_exit (fun () -> Printf.printf "\r%*s\r" (String.length !displayed) "")
 end
 
@@ -177,20 +192,29 @@ let split_lines s =
   loop ~acc:[] 0 0 ~last_is_cr:false
 ;;
 
-(* copy a file - fails if the file exists *)
-let copy ~header directive a b =
+let do_then_copy ~f a b =
   if Sys.file_exists b then fatal "%s already exists" b;
   let ic = open_in_bin a in
   let len = in_channel_length ic in
   let s = really_input_string ic len in
   close_in ic;
   let oc = open_out_bin b in
-  output_string oc header;
-  fprintf oc "#%s 1 %S\n" directive a;
+  f oc;
   output_string oc s;
   close_out oc
 ;;
 
+(* copy a file - fails if the file exists *)
+let copy a b = do_then_copy ~f:(fun _ -> ()) a b
+
+(* copy a file and insert a header - fails if the file exists *)
+let copy_with_header ~header a b = do_then_copy ~f:(fun oc -> output_string oc header) a b
+
+(* copy a file and insert a directive - fails if the file exists *)
+let copy_with_directive ~directive a b =
+  do_then_copy ~f:(fun oc -> fprintf oc "#%s 1 %S\n" directive a) a b
+;;
+
 let path_sep = if Sys.win32 then ';' else ':'
 
 let split_path s =
@@ -511,6 +535,7 @@ end = struct
           | Some x -> sprintf "cd %s && %s" x cmdline
           | None -> cmdline
         in
+        Status_line.clear ();
         prerr_endline cmdline;
         prerr_string stderr_s;
         flush stderr);
@@ -750,6 +775,13 @@ end
 
 module Library = struct
   module File_kind = struct
+    type asm =
+      { syntax : [ `Gas | `Intel ]
+      ; arch : [ `Amd64 ] option
+      ; os : [ `Win | `Unix ] option
+      ; assembler : [ `C_comp | `Msvc_asm ]
+      }
+
     type c =
       { arch : [ `Arm64 | `X86 ] option
       ; flags : string list
@@ -758,6 +790,7 @@ module Library = struct
     type t =
       | Header
       | C of c
+      | Asm of asm
       | Ml
       | Mli
       | Mll
@@ -770,6 +803,20 @@ module Library = struct
         | Not_found -> String.length fn
       in
       match String.sub fn ~pos:i ~len:(String.length fn - i) with
+      | (".S" | ".asm") as ext ->
+        let syntax = if ext = ".S" then `Gas else `Intel in
+        let os, arch, assembler =
+          let fn = Filename.remove_extension fn in
+          let check suffix = String.is_suffix fn ~suffix in
+          if check "x86-64_unix"
+          then Some `Unix, Some `Amd64, `C_comp
+          else if check "x86-64_windows_gnu"
+          then Some `Win, Some `Amd64, `C_comp
+          else if check "x86-64_windows_msvc"
+          then Some `Win, Some `Amd64, `Msvc_asm
+          else None, None, `C_comp
+        in
+        Some (Asm { syntax; arch; os; assembler })
       | ".c" ->
         let arch, flags =
           let fn = Filename.remove_extension fn in
@@ -785,7 +832,7 @@ module Library = struct
           else if check "_avx2"
           then x86 [ "-mavx2" ] [ "/arch:AVX2" ]
           else if check "_avx512"
-          then x86 [ "-mavx512f"; "-mavx512vl" ] [ "/arch:AVX512" ]
+          then x86 [ "-mavx512f"; "-mavx512vl"; "-mavx512bw" ] [ "/arch:AVX512" ]
           else if String.is_suffix fn ~suffix:"_neon"
           then Some `Arm64, []
           else None, []
@@ -814,6 +861,12 @@ module Library = struct
     ; flags : string list
     }
 
+  type asm_file =
+    { assembler : [ `C_comp | `Msvc_asm ]
+    ; flags : string list
+    ; out_file : string
+    }
+
   module Wrapper = struct
     type t =
       { toplevel_module : string
@@ -835,7 +888,7 @@ module Library = struct
     let mangle_filename t ({ file; kind } : source) =
       let base = Filename.basename file in
       match kind with
-      | C _ | Header -> base
+      | Asm _ | C _ | Header -> base
       | Mll | Mly | Ml | Mli ->
         let ext =
           match kind with
@@ -900,8 +953,26 @@ module Library = struct
     { ocaml_files : string list
     ; alias_file : string option
     ; c_files : c_file list
+    ; asm_files : asm_file list
     }
 
+  let keep_asm { File_kind.syntax; arch; os; assembler = _ } ~ccomp_type ~architecture =
+    (match os with
+     | Some `Unix -> Sys.os_type = "Unix"
+     | Some `Win -> Sys.os_type = "Win32"
+     | None -> true)
+    && (match syntax, ccomp_type with
+        | `Intel, "msvc" -> true
+        | `Gas, "msvc" -> false
+        | `Gas, _ -> true
+        | `Intel, _ -> false)
+    &&
+    match arch, architecture with
+    | None, _ -> true
+    | Some `Amd64, "amd64" -> true
+    | Some `Amd64, _ -> false
+  ;;
+
   let keep_c { File_kind.arch; flags = _ } ~architecture =
     match arch with
     | None -> true
@@ -922,7 +993,7 @@ module Library = struct
       let modules =
         List.fold_left files ~init:String.Set.empty ~f:(fun acc { file = fn; kind } ->
           match (kind : File_kind.t) with
-          | Header | C _ -> acc
+          | Asm _ | Header | C _ -> acc
           | Ml | Mli | Mll | Mly ->
             let module_name =
               let fn = Filename.basename fn in
@@ -942,11 +1013,14 @@ module Library = struct
            let mangled = Wrapper.mangle_filename wrapper source in
            let dst = build_dir ^/ mangled in
            (match kind with
+            | Asm _ ->
+              copy fn dst;
+              Fiber.return [ mangled ]
             | Header | C _ ->
-              copy ~header:"" "line" fn dst;
+              copy_with_directive ~directive:"line" fn dst;
               Fiber.return [ mangled ]
             | Ml | Mli ->
-              copy "" fn dst ~header;
+              copy_with_header ~header fn dst;
               Fiber.return [ mangled ]
             | Mll -> copy_lexer fn dst ~header >>> Fiber.return [ mangled ]
             | Mly ->
@@ -970,22 +1044,56 @@ module Library = struct
            Some (src, mangled))
     >>| fun (files, build_info_file) ->
     let alias_file = Wrapper.generate_wrapper wrapper modules in
-    let c_files, ocaml_files =
+    let c_files, ocaml_files, asm_files =
       let files =
         let files = List.concat files in
         match build_info_file with
         | None -> files
         | Some fn -> fn :: files
       in
+      let ext_obj =
+        try String.Map.find "ext_obj" ocaml_config with
+        | Not_found -> ".o"
+      in
+      let ccomp_type = String.Map.find "ccomp_type" ocaml_config in
       let architecture = String.Map.find "architecture" ocaml_config in
       List.partition_map_skip files ~f:(fun (src, fn) ->
         match src.kind with
         | C c ->
-          if keep_c c ~architecture then `Left { flags = c.flags; name = fn } else `Skip
-        | Ml | Mli | Mly | Mll -> `Right fn
-        | Header -> `Skip)
+          if keep_c c ~architecture
+          then (
+            let extra_flags =
+              if String.is_prefix ~prefix:"blake3_" fn
+              then (
+                match architecture with
+                | "x86" | "i386" | "i486" | "i586" | "i686" ->
+                  [ "-DBLAKE3_NO_SSE2"
+                  ; "-DBLAKE3_NO_SSE41"
+                  ; "-DBLAKE3_NO_AVX2"
+                  ; "-DBLAKE3_NO_AVX512"
+                  ]
+                | _ -> [])
+              else []
+            in
+            `Left { flags = extra_flags @ c.flags; name = fn })
+          else `Skip
+        | Ml | Mli | Mly | Mll -> `Middle fn
+        | Header -> `Skip
+        | Asm asm ->
+          if keep_asm asm ~ccomp_type ~architecture
+          then (
+            let out_file = Filename.chop_extension fn ^ ext_obj in
+            `Right
+              { flags =
+                  (match asm.assembler with
+                   | `C_comp -> [ "-c"; fn; "-o"; out_file ]
+                   | `Msvc_asm -> [ "/nologo"; "/quiet"; "/Fo" ^ out_file; "/c"; fn ])
+              ; assembler = asm.assembler
+              ; out_file
+              })
+          else `Skip)
     in
-    { ocaml_files; alias_file; c_files }
+    { ocaml_files; alias_file; c_files; asm_files }
   ;;
 end
 
@@ -1137,7 +1245,7 @@ let sort_files dependencies ~main =
 
 let common_build_args name ~external_includes ~external_libraries =
   List.concat
-    [ [ "-o"; Filename.concat "../_boot" (name ^ ".exe"); "-g" ]
+    [ [ "-o"; name ^ ".exe"; "-g" ]
     ; (match Config.mode with
        | Byte -> [ Config.output_complete_obj_arg ]
        | Native -> [])
@@ -1148,14 +1256,42 @@ let common_build_args name ~external_includes ~external_libraries =
 
 let allow_unstable_sources = [ "-alert"; "-unstable" ]
 
+let ocaml_warnings =
+  let warnings =
+    [ (* Warning 49 [no-cmi-file]: no cmi file was found in path for module *)
+      "-49"
+    ; (* Warning 23: all the fields are explicitly listed in this record: the
+        'with' clause is useless. 
+
+         In order to stay version independent, we use a trick with `with` by
+         creating a dummy value and filling in the fields available in every
+         OCaml version. forced_major_collections is the one missing in versions
+         older than 4.12. We therefore disable warning 23 for our purposes. *)
+      "-23"
+    ; (* Warning 53 [misplaced-attribute]: the "alert" attribute cannot appear
+        in this context
+
+        Any .mli files that begin wtih [@@@alert] will cause the compiler to
+        emit warning 53 due to the way we alias modules with an `open!`. It may
+        be possible to use the command line flag `-open` instead, however this
+        complicates dependency tracking so disabling this warning instead is
+        suitable for our purposes. *)
+      "-53"
+    ]
+  in
+  [ "-w"; String.concat ~sep:"" warnings ]
+;;
+
 let build
       ~ocaml_config
       ~dependencies
       ~c_files
+      ~asm_files
       ~build_flags
       ~link_flags
       { target = name, main; external_libraries; _ }
   =
+  let c_compiler = String.Map.find "c_compiler" ocaml_config in
   let ext_obj =
     try String.Map.find "ext_obj" ocaml_config with
     | Not_found -> ".o"
@@ -1188,7 +1324,8 @@ let build
              ~cwd:build_dir
              Config.compiler
              (List.concat
-                [ [ "-c"; "-g"; "-no-alias-deps"; "-w"; "-49-6" ]
+                [ [ "-c"; "-g"; "-no-alias-deps" ]
+                ; ocaml_warnings
                 ; allow_unstable_sources
                 ; external_includes
                 ; [ file ]
@@ -1196,14 +1333,27 @@ let build
   Fiber.fork_and_join_unit
     (fun () -> build (Filename.basename main))
     (fun () ->
-       Fiber.parallel_map c_files ~f:(fun { Library.name = file; flags } ->
-         let flags = List.map flags ~f:(fun flag -> [ "-ccopt"; flag ]) |> List.concat in
-         Process.run
-           ~cwd:build_dir
-           Config.compiler
-           (List.concat
-              [ [ "-c"; "-g" ]; external_includes; build_flags; [ file ]; flags ])
-         >>| fun () -> Filename.chop_extension file ^ ext_obj))
+       (Fiber.fork_and_join (fun () ->
+          Fiber.parallel_map c_files ~f:(fun { Library.name = file; flags } ->
+            let flags =
+              List.map flags ~f:(fun flag -> [ "-ccopt"; flag ]) |> List.concat
+            in
+            Process.run
+              ~cwd:build_dir
+              Config.compiler
+              (List.concat
+                 [ [ "-c"; "-g" ]; external_includes; build_flags; [ file ]; flags ])
+            >>| fun () -> Filename.chop_extension file ^ ext_obj)))
+         (fun () ->
+            Fiber.parallel_map asm_files ~f:(fun { Library.assembler; flags; out_file } ->
+              Process.run
+                ~cwd:build_dir
+                (match assembler with
+                 | `C_comp -> c_compiler
+                 | `Msvc_asm -> "ml64.exe")
+                flags
+              >>| fun () -> out_file))
+       >>| fun (x, y) -> x @ y)
   >>= fun obj_files ->
   let compiled_ml_files =
     let compiled_ml_ext =
@@ -1259,6 +1409,9 @@ let main () =
   let c_files =
     List.map ~f:(fun (lib : Library.t) -> lib.c_files) libraries |> List.concat
   in
+  let asm_files =
+    List.map ~f:(fun (lib : Library.t) -> lib.asm_files) libraries |> List.concat
+  in
   get_dependencies libraries
   >>= fun dependencies ->
   let ocaml_system =
@@ -1268,7 +1421,7 @@ let main () =
   in
   let build_flags = get_flags ocaml_system Libs.build_flags in
   let link_flags = get_flags ocaml_system Libs.link_flags in
-  build ~ocaml_config ~dependencies ~c_files ~build_flags ~link_flags task
+  build ~ocaml_config ~dependencies ~asm_files ~c_files ~build_flags ~link_flags task
 ;;
 
 let () = Fiber.run (main ())
