@@ -1,13 +1,13 @@
-open Stdune
+open Import
 
 type t
 
 include Comparable_intf.S with type key := t
-include Dune_sexp.Conv.S with type t := t
+include Conv.S with type t := t
 module Infix : Comparator.OPS with type t = t
 include Dune_util.Stringlike with type t := t
 
-val dune_site_syntax : Dune_sexp.Syntax.t
+val dune_site_syntax : Syntax.t
 
 module Modulelike (S : sig
     type t
