@@ -20,11 +20,13 @@ let compute_env ~mode =
             let+ v = Expander.eval_blang expander enabled_if in
             Some (Blang.Const v)
         in
+        let runtest_alias = Option.first_some local.runtest_alias parent.runtest_alias in
+        Option.iter ~f:Alias0.register_as_standard runtest_alias;
         { Js_of_ocaml.Env.compilation_mode =
             Option.first_some local.compilation_mode parent.compilation_mode
         ; sourcemap = Option.first_some local.sourcemap parent.sourcemap
         ; enabled_if
-        ; runtest_alias = Option.first_some local.runtest_alias parent.runtest_alias
+        ; runtest_alias
         ; flags =
             Js_of_ocaml.Flags.make
               ~spec:local.flags
@@ -314,6 +316,15 @@ let wasmoo ~dir sctx =
     "wasm_of_ocaml"
 ;;
 
+let jsoo_has_shapes jsoo_version =
+  match jsoo_version with
+  | Some version ->
+    (match Version.compare version (6, 1) with
+     | Lt -> false
+     | Gt | Eq -> true)
+  | None -> false
+;;
+
 type sub_command =
   | Compile
   | Link
@@ -520,6 +531,12 @@ let jsoo_archives ~mode ctx config lib =
            ]))
 ;;
 
+let cmo_js_of_module ~mode m =
+  Module_name.Unique.artifact_filename
+    (Module.obj_name m)
+    ~ext:(Js_of_ocaml.Ext.cmo ~mode)
+;;
+
 let link_rule
       ~mode
       cc
@@ -535,11 +552,6 @@ let link_rule
   =
   let sctx = Compilation_context.super_context cc in
   let dir = Compilation_context.dir cc in
-  let mod_name m =
-    Module_name.Unique.artifact_filename
-      (Module.obj_name m)
-      ~ext:(Js_of_ocaml.Ext.cmo ~mode)
-  in
   let ctx = Super_context.context sctx |> Context.build_context in
   let get_all =
     let open Action_builder.O in
@@ -565,12 +577,13 @@ let link_rule
     let special_units =
       List.concat_map to_link ~f:(function
         | Lib_flags.Lib_and_module.Lib _lib -> []
-        | Module (obj_dir, m) -> [ in_obj_dir' ~obj_dir ~config:None [ mod_name m ] ])
+        | Module (obj_dir, m) ->
+          [ in_obj_dir' ~obj_dir ~config:None [ cmo_js_of_module ~mode m ] ])
     in
     let all_libs = List.concat_map libs ~f:(jsoo_archives ~mode ctx config) in
     let all_other_modules =
       List.map cm ~f:(fun m ->
-        Path.build (in_obj_dir ~obj_dir ~config:None [ mod_name m ]))
+        Path.build (in_obj_dir ~obj_dir ~config:None [ cmo_js_of_module ~mode m ]))
     in
     let std_exit =
       Path.build
@@ -604,8 +617,25 @@ let link_rule
     ~sourcemap
 ;;
 
-let build_cm' sctx ~dir ~in_context ~mode ~src ~target ~config ~sourcemap =
-  let spec = Command.Args.Dep src in
+let build_cm' sctx ~dir ~in_context ~mode ~src ~target ~config ~shapes ~sourcemap =
+  let spec =
+    Command.Args.(
+      S
+        [ Dep src
+        ; Dyn
+            (let open Action_builder.O in
+             let* jsoo_version =
+               let* jsoo = jsoo ~dir sctx in
+               Action_builder.of_memo @@ Version.jsoo_version jsoo
+             in
+             let+ shapes =
+               match jsoo_has_shapes jsoo_version with
+               | false -> Action_builder.return []
+               | true -> shapes
+             in
+             S (List.map shapes ~f:(fun s -> S [ A "--load-shape"; Dep s ])))
+        ])
+  in
   let flags = in_context.Js_of_ocaml.In_context.flags in
   js_of_ocaml_rule
     sctx
@@ -620,9 +650,29 @@ let build_cm' sctx ~dir ~in_context ~mode ~src ~target ~config ~sourcemap =
     ~sourcemap
 ;;
 
-let build_cm sctx ~dir ~in_context ~mode ~src ~obj_dir ~config =
+let build_cm cctx ~dir ~in_context ~mode ~src ~obj_dir ~deps ~config:config_opt =
   let name = with_js_ext ~mode (Path.basename src) in
-  let target = in_obj_dir ~obj_dir ~config [ name ] in
+  let target = in_obj_dir ~obj_dir ~config:config_opt [ name ] in
+  let sctx = Compilation_context.super_context cctx in
+  let ctx = Super_context.context sctx |> Context.build_context in
+  let shapes =
+    let open Action_builder.O in
+    let+ libs = Resolve.Memo.read (Compilation_context.requires_link cctx)
+    and+ deps = deps
+    and+ config =
+      match config_opt with
+      | None ->
+        let flags = in_context.Js_of_ocaml.In_context.flags in
+        js_of_ocaml_flags sctx ~dir ~mode flags
+        |> Action_builder.bind ~f:(fun (x : _ Js_of_ocaml.Flags.t) -> x.compile)
+        |> Action_builder.map ~f:Config.of_flags
+      | Some config -> Action_builder.return config
+    in
+    (Path.build (in_build_dir ctx ~config [ "stdlib"; with_js_ext ~mode "stdlib.cma" ])
+     :: List.concat_map libs ~f:(fun lib -> jsoo_archives ~mode ctx config lib))
+    @ List.map deps ~f:(fun m ->
+      Path.build (in_obj_dir ~obj_dir ~config:config_opt [ cmo_js_of_module ~mode m ]))
+  in
   build_cm'
     sctx
     ~dir
@@ -630,7 +680,8 @@ let build_cm sctx ~dir ~in_context ~mode ~src ~obj_dir ~config =
     ~mode
     ~src
     ~target
-    ~config:(Option.map config ~f:Action_builder.return)
+    ~shapes
+    ~config:(Option.map config_opt ~f:Action_builder.return)
     ~sourcemap:Js_of_ocaml.Sourcemap.Inline
 ;;
 
@@ -647,6 +698,11 @@ let setup_separate_compilation_rules sctx components =
      | None -> Memo.return ()
      | Some pkg ->
        let info = Lib.info pkg in
+       let requires =
+         let open Resolve.Memo.O in
+         let* reqs = Lib.requires pkg in
+         Lib.closure ~linking:false reqs
+       in
        let lib_name = Lib_name.to_string (Lib.name pkg) in
        let* archives =
          let archives = (Lib_info.archives info).byte in
@@ -677,6 +733,23 @@ let setup_separate_compilation_rules sctx components =
            let target =
              in_build_dir build_context ~config [ lib_name; with_js_ext ~mode name ]
            in
+           let shapes =
+             let open Action_builder.O in
+             let+ requires = Resolve.Memo.read requires in
+             let l =
+               List.concat_map requires ~f:(fun lib ->
+                 jsoo_archives ~mode build_context config lib)
+             in
+             match lib_name with
+             | "stdlib" -> l
+             | _ ->
+               Path.build
+                 (in_build_dir
+                    build_context
+                    ~config
+                    [ "stdlib"; with_js_ext ~mode "stdlib.cma" ])
+               :: l
+           in
            build_cm'
              sctx
              ~dir
@@ -686,6 +759,7 @@ let setup_separate_compilation_rules sctx components =
              ~target
              ~config:(Some (Action_builder.return config))
              ~sourcemap:Js_of_ocaml.Sourcemap.Inline
+             ~shapes
            |> Super_context.add_rule sctx ~dir)))
 ;;
 
