@@ -17,8 +17,7 @@ let find_module sctx src =
   let src =
     Path.Build.append_source (Context.build_dir (Super_context.context sctx)) src
   in
-  let dir = Path.Build.parent_exn src in
-  let module_name =
+  match
     let open Option.O in
     let* name =
       let fname = Path.Build.basename src in
@@ -26,34 +25,30 @@ let find_module sctx src =
       if String.equal fname name then None else Some name
     in
     Module_name.of_string_opt name
-  in
-  match module_name with
+  with
   | None -> Memo.return None
   | Some module_name ->
+    let dir = Path.Build.parent_exn src in
     let* dir_contents = drop_rules @@ fun () -> Dir_contents.get sctx ~dir in
-    let* ocaml = Dir_contents.ocaml dir_contents in
-    let stanza =
-      match Ml_sources.find_origin ocaml [ module_name ] with
-      | Some (Executables exes) -> Some (`Executables exes)
-      | Some (Library lib) -> Some (`Library lib)
-      | None | Some (Melange _) -> None
-    in
-    (match stanza with
+    let* ocaml = Dir_contents.ocaml dir_contents
+    and* scope = Dir_contents.dir dir_contents |> Scope.DB.find_by_dir in
+    Ml_sources.find_origin ocaml ~libs:(Scope.libs scope) [ module_name ]
+    >>= (function
      | None -> Memo.return None
-     | Some stanza ->
-       let* scope = Scope.DB.find_by_dir dir in
+     | Some origin ->
        let* expander = Super_context.expander sctx ~dir in
        let+ cctx, merlin =
          drop_rules
          @@ fun () ->
-         match stanza with
-         | `Executables exes ->
-           Exe_rules.rules ~sctx ~dir ~dir_contents ~scope ~expander exes
-         | `Library lib -> Lib_rules.rules lib ~sctx ~dir_contents ~dir ~expander ~scope
+         match origin with
+         | Executables exes -> Exe_rules.rules ~sctx ~dir_contents ~scope ~expander exes
+         | Library lib -> Lib_rules.rules lib ~sctx ~dir_contents ~expander ~scope
+         | Melange mel ->
+           Melange_rules.setup_emit_cmj_rules ~sctx ~dir_contents ~expander ~scope mel
        in
-       let modules = Compilation_context.modules cctx in
        let module_ =
-         match Modules.find modules module_name with
+         let modules = Compilation_context.modules cctx in
+         match Modules.With_vlib.find modules module_name with
          | Some m -> m
          | None ->
            User_error.raise
@@ -62,41 +57,40 @@ let find_module sctx src =
                  (Path.Build.to_string_maybe_quoted src)
              ]
        in
-       Some (module_, cctx, merlin))
+       Some (module_, cctx, merlin, origin))
 ;;
 
 let module_deps cctx module_ =
-  let dep_graph =
-    let dg = Compilation_context.dep_graphs cctx in
-    Ocaml.Ml_kind.Dict.get dg Impl
+  let+ graph, _ =
+    let dep_graph =
+      let dg = Compilation_context.dep_graphs cctx in
+      Ocaml.Ml_kind.Dict.get dg Impl
+    in
+    Dep_graph.deps_of dep_graph module_ |> Action_builder.evaluate_and_collect_facts
   in
-  let action = Dep_graph.deps_of dep_graph module_ in
-  let+ graph, _ = Action_builder.run action Eager in
   graph
 ;;
 
 let gen_rules sctx ~dir:rules_dir ~comps =
   let src = Path.Source.L.relative Path.Source.root comps in
-  let* mod_ = find_module sctx src in
-  match mod_ with
+  find_module sctx src
+  >>= function
   | None -> Memo.return ()
-  | Some (module_, cctx, _merlin) ->
+  | Some (module_, cctx, _merlin, _) ->
     let module_ = Module.set_source module_ Intf None in
     let* () =
-      let* module_deps = module_deps cctx module_ in
-      let files =
-        let obj_dir = Compilation_context.obj_dir cctx in
-        List.filter_map module_deps ~f:(fun module_ ->
-          Obj_dir.Module.cm_file obj_dir module_ ~kind:(Ocaml Cmi))
-      in
-      Memo.parallel_iter files ~f:(fun file ->
-        let src = Path.build file in
-        let dst = Path.Build.relative rules_dir (Path.Build.basename file) in
-        Super_context.add_rule sctx ~dir:rules_dir (Action_builder.symlink ~src ~dst))
+      (let obj_dir = Compilation_context.obj_dir cctx in
+       module_deps cctx module_
+       >>| List.filter_map ~f:(Obj_dir.Module.cm_file obj_dir ~kind:(Ocaml Cmi)))
+      >>= Memo.parallel_iter ~f:(fun file ->
+        (let src = Path.build file in
+         let dst = Path.Build.relative rules_dir (Path.Build.basename file) in
+         Action_builder.symlink ~src ~dst)
+        |> Super_context.add_rule sctx ~dir:rules_dir)
     in
     let cctx =
-      let obj_dir = private_obj_dir (Super_context.context sctx) src in
-      Compilation_context.set_obj_dir cctx obj_dir
+      private_obj_dir (Super_context.context sctx) src
+      |> Compilation_context.set_obj_dir cctx
       |> Compilation_context.without_bin_annot
       |> Compilation_context.set_modes
            ~modes:
